
/*

Pitchbender for PC

Andreja Andric 2018

The purpose of the Pitchbender program is threefold:
1. to resample the input at a sample rate that is a given multiple or
fraction of the original sample rate,
2. to be configured to follow any number of such fractions one after
the other, and to change the sample rate accordingly at precisely the
right moments, and
3. to do all the above processing and generate the output in real time
and without any noticeable delay
The selection of the input files, and other options can be configured 
in the file pitchbenderconfig.txt.
The program uses an ASIO driver, and you have to have one, otherwise 
it will not work.
Pitchbender uses internal buffering to allow the sample flow to slow
down and speed up at will. Obviously, it can go faster than real speed
only if it had slowed down before, because it would be impossible to
go absolutely faster than the real time. However, if it happens that
the configuration file is set up so that Pitchbender has to go
before its time, a solution is provided in a sort of a loop  the
sample pointer goes a fixed distance back when it hits the last sample
from the soundcard input. This creates a convincing looping in high
register, but should nevertheless be avoided because it also creates
unpleasant clicks in the sound output. The input files should be
composed in such a way that this cannot happen  every higher note
should always be preceded by its lower counterpart, so after every two
subsequent notes the sample counter is exactly where it would be if
there were no pitch altering at all.
The input files for Pitchbender have simple structure: its a list of
time-pitch pairs each of which indicate for how long a given relative
pitch is to sound. The list is preceded by a "pivot" number that
indicates the value in respect of which the other relative pitches are
"relative". For example if this number is 1, then relative pitch of 0.5
indicates slowing down to a half and an octave lower pitch, relative
pitch of 2 indicates an octave transposition and twice faster playing,
and so on. Setting the file this way makes it easy to transpose an entire
input file just by varying this pivot number. Lists of time-pitch pairs
can be optionally followed by a single negative number, that indicates
the number of repeats of the preceding group (between this and the
previous repetition mark).

To compile it requires MS Visual Studio, portaudio library and ASIOSDK2 from Steinberg
(see instructions for building portaudio library with ASIO support on the portaudio webpage).
Build Portaudio for ASIO only.

I used Visual Studio 15 and portaudio with ASIO support from pa_stable_v190600_20161030.tar 
archive on Windows 10 to build the Portaudio library. 

After you have unpacked and compiled portaudio with ASIOSDK, compile Pitchbender with the 
following command in the Command window:

cl pitchbender.c .\pabuild\Release32\portaudio_x86.lib

I used Visual C++ 2010 Express to build the Pitchbender.

Deployment to another computer requires:
1) Installing ASIO4ALL driver from the ASIO4ALL website (Note: only v2.14 works with Windows 10)
2) copying pitchbender.exe, portaudio_x86.dll, pitchbenderconfig.txt and the directories with patches.

Edit pitchbenderconfig.txt to specify the input file(s), 
whether you wish the input file to be looped infinitely, default
device numbers for input and output, default key for controlling
the flow and the stereo channel handling (left-to-both, right-to-both, 
normal-stereo).

Upon start, the program also displays the ASIO driver panel. The buffer 
length has to be specified there. If this is modified, then the program 
has to be stopped and started again, because this and other relevant values 
received from the driver are picked at the program start.

Example config file:

# Pitchbender config file
# Format:
# field: value
# Allowed fields: File, InfiniteLoop, DefaultKeyCode, DefaultKeyName, Stereo, DefaultInput, DefaultOutput

File: 1_Introduction.txt
File: 2_Dance1.txt
File: 3_Intermission.txt
File: 4_Dance2.txt
File: 5_NightWalk.txt
InfiniteLoop: No
DefaultKeyCode: 32
DefaultKeyName: Space
Stereo: normal-stereo
DefaultInput: 0
DefaultOutput: 1

Run the program by invoking it from the Command window or by doubleclick.


 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "portaudio/include/portaudio.h"

/* #define SAMPLE_RATE  (17932) // Test failure to open with this value. */
#define SAMPLE_RATE       (44100)
/* #define DITHER_FLAG     (paDitherOff)  */
#define DITHER_FLAG           (0)

/* Select sample format. */
#define PA_SAMPLE_TYPE  paFloat32
#define SAMPLE_SIZE (4)
#define SAMPLE_SILENCE  (0.0f)
#define PRINTF_S_FORMAT "%.8f"
#define SAMPLE float

#include <conio.h>
#include <fcntl.h>
#include <math.h>
#include <memory.h>
#include <time.h>

#include "portaudio/include/pa_asio.h"

/* represents a single move of the sample pointer.
   move field tells how far the sample pointer will 
   move at he current instant. if it's zero, it won't
   move. If it's 1 it passes to the next. If it's 2,
   it'll skip the next one and pass to the one after
   it.
   In value field we keep the last value of k in the
   line algorithm, so we know when to close the loop.
*/
typedef struct samplemove
{
   int move;
   double value;
   int delete;
   struct samplemove* next;
} SAMPLEMOVE;
typedef SAMPLEMOVE* samplemoveptr;

/* a note is described with relative frequency and
   duration. Repeat field points back to the beginning
   of the phrase if this is the last note. Otherwise
   it's NULL. moves field points to the list of moves
   that implement the given frequency. Repetitions field
   keeps the total repetitions count, and iterations field
   keeps current repetition count.
*/
typedef struct note
{
   long n;
   float duration;
   float frequency;
   int repetitions;
   int iterations;
   struct note* repeat;
   struct note* next;
   samplemoveptr moves;
} NOTE;
typedef NOTE* noteptr;

/*
Main thing kept in channel is a list of notes. last_i
and last_k help implement the straight line algorithm
in constructing the list of sample pointer moves
*/
typedef struct channel
{
   int num;
   int ms_timer;
   long last_i;
   double last_k;
   int samples;
   int samples_per_ms;
   char *bigbuffer;
   char *inbuffer;
   char *outbuffer;
   noteptr melodyhead;
   noteptr melodytail;
   noteptr restart;
   noteptr currentnote;
   samplemoveptr lastmove;
} CHANNEL;
typedef CHANNEL* channelptr;

/* 
   These are provisory values, that will be overwritten
   on initialization of the library.
   Sample rate is fixed at 44100. As the smallest note
   duration is 1 millisecond, and 44100 is not divisible
   by 1000, for every 9 out of 10 milliseconds the sample
   count is 44, and for the remaining one it's 45.
   Sample size is 2 bytes.
   BufferNum is depth of the array of buffers. This gets
   configured as well in the initialization.
*/
int mainRate=44100;
int add_per_ms = 44;
int corr_per_ms = 10;
int SampleSize = 2;
double pivot;
int ChannelNum = 2;
int BufferNum = 200;
int BufferLen = 512;
channelptr *Channels=NULL;
channelptr CurrentChannel=NULL;
int cc=0;
int Initialized=0;
unsigned long callsCount;
int DelayMsgCnt=0;
int InfiniteLoop=0;

/*
Creating and initializing a channel.
*/
channelptr createChannel(int num)
{
   channelptr newchannel = (channelptr)malloc(sizeof(CHANNEL));
   if(newchannel==NULL)
   {
      printf("Out of memory.\n");
      return NULL;
   }
   newchannel->num = num;
   newchannel->ms_timer=0;
   newchannel->last_k=0;
   newchannel->samples=0;
   newchannel->samples_per_ms=45;
   newchannel->bigbuffer = (char*)malloc(BufferNum*BufferLen+1);
   if(newchannel->bigbuffer==NULL)
   {
      printf("Out of memory.\n");
      return NULL;
   }
   
   newchannel->inbuffer = newchannel->bigbuffer+(BufferNum-1)*BufferLen;
 
   newchannel->outbuffer = (char*)malloc(BufferLen+1);
   if(newchannel->outbuffer==NULL)
   {
      printf("Out of memory.\n");
      return NULL;
   }

   newchannel->inbuffer[BufferLen]='\0';
   newchannel->outbuffer[BufferLen]='\0';
   newchannel->melodyhead=NULL;
   newchannel->melodytail=NULL;
   newchannel->restart = NULL;
   newchannel->currentnote=NULL;
   newchannel->last_i=BufferLen;

   newchannel->lastmove = NULL;

   return newchannel;
}

/*
Copying a channel
*/
int CopyChannel(channelptr src, channelptr dst)
{
   if(src==NULL||dst==NULL)
   {
      printf("%s=0\n", src==NULL?"src":"dst");
      printf("%s=0\n", dst==NULL?"dst":"src");
      return 1;
   }
   dst->num = src->num+1;
   dst->ms_timer=0;
   dst->last_k=0;
   dst->samples=0;
   dst->samples_per_ms=45;
   dst->bigbuffer = (char*)malloc(BufferNum*BufferLen+1);
   if(dst->bigbuffer==NULL)
   {
      printf("Out of memory.\n");
      return 0;
   }
   
   dst->inbuffer = dst->bigbuffer+(BufferNum-1)*BufferLen;
 
   dst->outbuffer = (char*)malloc(BufferLen+1);
   if(dst->outbuffer==NULL)
   {
      printf("Out of memory.\n");
      return 0;
   }

   dst->inbuffer[BufferLen]='\0';
   dst->outbuffer[BufferLen]='\0';
   dst->melodyhead=src->melodyhead;
   dst->melodytail=src->melodytail;
   dst->restart = src->restart;
   dst->currentnote=src->restart;
   dst->last_i=src->last_i;
   dst->lastmove = src->lastmove;

   return 1;
}

/*
Adding a single sample pointer move
*/
samplemoveptr addmove(int move, double value, samplemoveptr tail, noteptr note)
{
   samplemoveptr currmove = (samplemoveptr)malloc(sizeof(SAMPLEMOVE));
   if(!currmove)
   {
      printf("Out of memory.\n");
      return NULL;
   }
   if(move<0)
   {
      printf("error: move < 0\n");
      return NULL;
   }
 
   currmove->move=move;
   currmove->value = value;
   currmove->delete = 0;
   currmove->next = NULL;
   if(note->moves==NULL)
      note->moves=currmove;
   else
      tail->next=currmove;
   return currmove;
}

/*
Relative pitch of each note will be implemented with
a finite number of sample pointer moves. You can imagine
it as a finite number of digits in a decimal number.
So it's necessary before every addition to check whether
the current value of k appears previously, or its near
approximation.
*/
samplemoveptr findmove(double k, samplemoveptr moves)
{
   samplemoveptr currmove;
   if(moves==NULL||moves->next==NULL)
      return NULL;
   currmove = moves->next;
   while(currmove)
   {
      if(abs((int)(currmove->value-k))<0.001)
         return currmove;
      currmove = currmove->next;
   }
   return NULL;
}

/*
An adaptation of the straight line algorithm for creating
all sample pointer moves that implement a relative pitch
of a note
*/
void nextmove(channelptr chan)
{
   samplemoveptr currmove = NULL;
   double old_k = chan->last_k;
   unsigned long old_i = chan->last_i;
   int loop=1;
   
   while(loop)
   {
      if(chan->last_k<0)
      {
         chan->last_k+=pivot;
         chan->last_i++;
      }
      else
      {
         chan->last_k-=chan->melodytail->frequency;
         loop=0;
      }
   }
   currmove = findmove(old_k, chan->melodytail->moves);
   if(currmove!=NULL)
   {
      chan->lastmove->next = currmove;
      chan->lastmove = NULL;
   }
   else
   {
      chan->lastmove = addmove(chan->last_i-old_i, old_k, chan->lastmove, chan->melodytail);
   }
}

/*
Adding a note to channel. From here we call the nextmove
function above
*/
noteptr addnote(float frequency, float duration, channelptr chan)
{
   samplemoveptr kill;
   noteptr currnote = (noteptr)malloc(sizeof(NOTE));
   if(!currnote)
   {
      printf("Out of memory.\n");
      return NULL;
   }
   if(frequency<=0||duration<=0)
   {
      printf("Zero or negative note in melody (f=%f, d=%f).\n", frequency, duration);
      return NULL;
   }
 
   currnote->frequency=frequency;
   currnote->duration=duration;
   currnote->repetitions=1;
   currnote->iterations = 0;
   currnote->next = NULL;
   currnote->repeat = NULL;
   currnote->moves = NULL;
   if(chan->melodyhead==NULL)
   {
      chan->melodyhead=currnote;
      chan->restart=currnote;
      currnote->n=1;
   }
   else
   {
      chan->melodytail->next=currnote;
      if(chan->melodytail->repeat!=NULL)
         chan->restart=currnote;
      currnote->n=chan->melodytail->n+1;
   }
   chan->melodytail = currnote;

   do
      nextmove(chan);
   while(chan->lastmove);

   chan->last_k=0; 
   chan->last_i=0;

   kill = currnote->moves;
   currnote->moves = currnote->moves->next;
   free(kill);

   return currnote;
}

/*
Checking the initialization values for coherence.
Summary: supported rates are 11025, 22050 and 44100;
at least one channel; at least one buffer; at least
1 byte per buffer; 1, 2 or 4 bytes per sample.
*/
int checkvalues(int channelnum, int bufnum, int buflen, int rate, int samplesize)
{
   if(rate!=11025&&rate!=22050&&rate!=44100)
   {
      printf("Supported rates: 11025, 22050 and 44100. You supplied %d\n", rate);
      return 1;
   }
   if(channelnum<1)
   {
      printf("At least 1 channel required.\n");
      return 1;
   }
   if(buflen<1)
   {
      printf("At least 1 byte required per buffer.\n");
      return 1;
   }
   if(bufnum<1)
   {
      printf("At least 1 buffer required.\n");
      return 1;
   }
   if(samplesize!=1&&samplesize!=2&&samplesize!=4&&samplesize!=8&&samplesize!=12&&samplesize!=16)
   {
      printf("Supported sample sizes are 1, 2, 4, 8, 12 and 16 bytes. You supplied %d\n", samplesize);
      return 1;
   }
   return 0;
}

/*
The list of pitch modification is "monophonic". This means
that there is only one channel with some real content.
But, as the soundcard can have any number of channels
to process with this single channel of notes, and this
requires multiple copies of sample pointers and the rest,
all channels save one are created as copies of that first
one. Only pointers and counters are copied, not the buffers
themselves.
*/
int CreateOtherChannels(channelptr* chans, int channelnum)
{
   int i;
   for(i=1;i<channelnum;i++)
   {
      chans[i]=createChannel(i);
      if(!chans[i])
         return 0;
      if(!CopyChannel(chans[0], chans[i]))
         return 0;
   }
   return 1;
}

void CleanUp()
{
   int i;
   noteptr kill, curr;
   if(Channels==NULL)
      return;
   curr = Channels[0]->melodyhead;
    if(InfiniteLoop)
    {
        curr = curr->next;
        Channels[0]->melodyhead->next = NULL;
    }
   kill = curr;
   while(curr!=NULL)
   {
      samplemoveptr currmove = curr->moves;
      samplemoveptr killmove = currmove;
      while(currmove!=NULL)
      {
         currmove->delete=1;
         if(currmove->next->delete==1)
            currmove->next = NULL;
         currmove = currmove->next;
      }
      currmove = killmove;
      while(currmove!=NULL)
      {
         currmove = currmove->next;
         free(killmove);
         killmove = currmove;
      }
      curr = curr->next;
      free(kill);
      kill = curr;
   }
   
   for(i=0;i<ChannelNum;i++)
   {
      free(Channels[i]->bigbuffer);
      free(Channels[i]->outbuffer);
      free(Channels[i]);
   }
   free(Channels);
   Channels = NULL;
   Initialized = 0;
   DelayMsgCnt = 0;
}

/*
Initialization of the library. The calling program passes:
-number of channels
-number of buffers
-buffer length
-sample rate
-sample size
-file name with the configuration file to read
-whether the session is to be saved or not

After checking all the values, the library tries to load
a configuration file and create corresponding data 
structures in memory.
The configuration files for pitchbender have simple 
structure: its a list of time-pitch pairs each of 
which indicate for how long a given relative pitch is 
to sound. The list is preceded by a pivot number 
that indicates the value in respect of which the other 
relative pitches are relative. For example if this 
number is 1, then relative pitch of 0.5 indicates 
slowing down to a half and an octave lower pitch, 
relative pitch of 2 indicates an octave transposition 
and twice faster playing, and so on. Setting the file 
this way makes it easy to transpose an entire 
configuration file just by varying this pivot number. 
Lists of time-pitch pairs can be optionally followed 
by a single negative number, that indicates the number 
of repeats of the preceding group (between this and 
the previous repetition mark). 

If the session is to be recorded, the output file is
also opened.

If all went well, the library is initialized.

*/

int PitchBenderInitialize(int channelnum, int bufnum, int buflen, int rate, int samplesize, char* filename, int infLoop)
{
   int i;
   FILE* in;
   if(checkvalues(channelnum, bufnum, buflen, rate, samplesize))
      return 1;

   CleanUp();

   BufferNum = bufnum;
   BufferLen = buflen;
   SampleSize = samplesize;
   ChannelNum = channelnum;

   Channels = (channelptr*)malloc(channelnum*sizeof(channelptr));
   if(Channels==NULL)
   {
      printf("Out of memory.\n");
      return 1;
   }
   Channels[0]=NULL;

   printf("Opening file %s\n", filename);
   in=fopen(filename, "rt");
   if(!in)
   {
      printf("Can't open %s\n", filename);
      return 1;
   }
   fscanf(in, "%lf", &pivot);

   for(;;)
   {
      double d, f;
      char inputstring[100];
      int repcount, repIndication=0;
      fscanf(in, "%s", inputstring);
      if(sscanf(inputstring, "%lf", &d)!=1)
      {
         printf("Invalid input string %s\n", inputstring);
         fclose(in);
         return 1;
      }
      if(feof(in))
         break;
      if(d<0)
      {
         repIndication=1;
         repcount=-(int)d;
      }
      else
      if(d==0)
      {
         printf("Invalid input string %s\n", inputstring);
         fclose(in);
         return 1;
      }
      else
      {
         repIndication=0;
         fscanf(in, "%s", inputstring);
         if(sscanf(inputstring, "%lf", &f)!=1)
         {
            printf("Invalid input string %s\n", inputstring);
            fclose(in);
            return 1;
         }
      }

      if(repIndication==1)
      {
         if(repcount<1)
         {
            printf("Invalid repetition count %d.\n", repcount);
            fclose(in);
            return 1;
         }
         else
         if(Channels[0]==NULL)
         {
            printf("Misplaced repetition count %d.\n", repcount);
            fclose(in);
            return 1;
         }
         Channels[0]->melodytail->repetitions=repcount-1;
         Channels[0]->melodytail->repeat=Channels[0]->restart;
      }
      else
      {
         if(Channels[0]==NULL)
            Channels[0] = createChannel(0);

         if(!addnote(f, d, Channels[0]))
         {
            fclose(in);
            return 1;
         }
      }
   }
   fclose(in);

   /* we close the circle in the end */
   if(Channels[0]->melodytail!=NULL)
   {
      if(infLoop)
         Channels[0]->melodytail->next=Channels[0]->melodyhead;
      else
         Channels[0]->melodytail->next=NULL;
   }
   else
   {
      printf("Channel empty.\n");
      return 1;
   }

   add_per_ms = add_per_ms * rate / mainRate;
   corr_per_ms = corr_per_ms * mainRate/rate;

   CurrentChannel = Channels[0];
   if(!CreateOtherChannels(Channels, ChannelNum))
      return 1;

   cc=0;
   callsCount = 0;

   printf("Initialized, baby.\n");
   Initialized = 1;
   return 0;
}

/*
This accepts a data buffer at a time, processes it
according to the current position in the note list,
and gives back the result.
*/
int PitchBend(const void* sndcardin, const void* sndcardout)
{
   int i, j, testcnt=0;
   noteptr note;

   if(!Initialized)
   {
      printf("Effect not initialized.\n");
      return 1;
   }

   CurrentChannel=Channels[cc];
   cc++;
   if(cc>=ChannelNum)
      cc=0;

   if(CurrentChannel->melodyhead==NULL)
   {
      memcpy(sndcardout, sndcardin, BufferLen);
      return 0;
   }

   if(callsCount%1000==0)
   {
      printf("%d. Delay on channel %d = %ld samples\n", ++DelayMsgCnt,
         CurrentChannel->num, -CurrentChannel->last_i);
   }
   callsCount++;

   CurrentChannel->last_i-=BufferLen;
   memmove(CurrentChannel->inbuffer-(BufferNum-1)*BufferLen, CurrentChannel->inbuffer-(BufferNum-2)*BufferLen, (BufferNum-1)*BufferLen);
   memcpy(CurrentChannel->inbuffer, sndcardin, BufferLen);

   note = CurrentChannel->melodyhead;
   for(i=0;i<BufferLen/SampleSize&&CurrentChannel->melodyhead!=NULL;i++)
   {
      if(CurrentChannel->last_i>=BufferLen)
      {
         CurrentChannel->last_i-=BufferLen;
      }
      if(CurrentChannel->inbuffer+CurrentChannel->last_i<CurrentChannel->bigbuffer)
         CurrentChannel->last_i+=BufferLen;
      for(j=0;j<SampleSize;j++)
         *(CurrentChannel->outbuffer+SampleSize*i+j)=*(CurrentChannel->inbuffer+CurrentChannel->last_i+j);
      CurrentChannel->last_i=CurrentChannel->last_i+SampleSize*CurrentChannel->melodyhead->moves->move;
      CurrentChannel->melodyhead->moves = CurrentChannel->melodyhead->moves->next;
      CurrentChannel->samples++;
      if(CurrentChannel->samples==CurrentChannel->samples_per_ms)
      {
         CurrentChannel->samples=0;
         CurrentChannel->ms_timer++;
         CurrentChannel->samples_per_ms=add_per_ms+(CurrentChannel->ms_timer%corr_per_ms==0);
         if(CurrentChannel->ms_timer==CurrentChannel->melodyhead->duration)
         {
            CurrentChannel->ms_timer=0;
            if(CurrentChannel->melodyhead->repeat!=NULL)
            {
               if(CurrentChannel->melodyhead->iterations<CurrentChannel->melodyhead->repetitions)
               {
                  CurrentChannel->melodyhead->iterations++;
                  CurrentChannel->melodyhead=CurrentChannel->melodyhead->repeat;
               }
               else
               {
                  CurrentChannel->melodyhead->iterations=0;
                  CurrentChannel->melodyhead=CurrentChannel->melodyhead->next;
               }
            }
            else
            {
               CurrentChannel->melodyhead=CurrentChannel->melodyhead->next;
            }
         }
      }
   }

   memcpy(sndcardout, CurrentChannel->outbuffer, BufferLen);
   return 0;
}

void PitchBenderClose()
{
   CleanUp();
}

int ChoiceWithDefault(char keyBottomChoice, char keyTopChoice, char keyDefault, char *keyDefaultName, int valDefault)
{
    char ch;
    printf("Press '%c' - '%c' to choose, or %s for default ('%d')\n", 
       keyBottomChoice, keyTopChoice, keyDefaultName, valDefault);
    while( 1 )
    {
       ch = getch();
       if((ch>=keyBottomChoice&&ch<=keyTopChoice)||ch==keyDefault)
          break;
    }
    printf("Key pressed: '%c' (ASCII code: '%d')\n", ch, ch);
    if(ch==keyDefault)
       return valDefault;
    else
       return (ch-keyBottomChoice);
}

int ChoiceFromStringWithDefault(char* ChoiceString, char keyDefault, char *keyDefaultName, int valDefault)
{
    char ch;
    int i, len = strlen(ChoiceString);
    printf("Press any of '");
    for(i=0;i<len;i++)
        printf("%c%s", ChoiceString[i], i==len-1?"'":"', '");
    printf(" to choose, or %s for default ('%d')\n", keyDefaultName, valDefault);
    while( 1 )
    {
       ch = getch();
       if(NULL!=strchr(ChoiceString, ch)||ch==keyDefault)
          break;
       if(ch!=-1)
          printf("Key code: %d\n", ch);
    }
    printf("Key pressed: '%c' (ASCII code: '%d')\n", ch, ch);
    if(ch==keyDefault)
       return valDefault;
    else
       return (ch-'0');
}

int WaitForKeyPressed (char keyToProceed, char *keyNameToProceed, char keySpecial, char *specialAction)
{
    char ch;
    printf("Press %s to proceed, '%c' to %s\n", keyNameToProceed, keySpecial, specialAction);
    while( 1 )
    {
       ch = getch();
       if(ch==keyToProceed||ch==keySpecial)
          break;
       if(ch!=-1)
          printf("Key code: %d\n", ch);
    }
    return (ch==keySpecial);
}

void WaitForAnyKeyPressed ()
{
    char ch;
    printf("Press any key to proceed\n");
    ch = getch();
}

static int gNumNoInputs = 0;
static int StereoMode = 0;
int bufferLenDiscrepancy = 0;

#define LINELEN 500

int FramesPerBuffer;
char **InputFiles;
char **inputfilecontent;
int InputFileCount;
int DefaultKeyCode=32;
char *DefaultKeyName;
int DefaultInput, DefaultOutput;
int DefaultInputPos, DefaultOutputPos;
int inputfilelines;

static int pitchbendCallback( const void *inputBuffer, void *outputBuffer,
                        unsigned long framesPerBuffer,
                        const PaStreamCallbackTimeInfo* timeInfo,
                        PaStreamCallbackFlags statusFlags,
                        void *userData )
{
    SAMPLE *out = (SAMPLE*)outputBuffer;
    const SAMPLE *in = (const SAMPLE*)inputBuffer;
    unsigned int i;
    int err=0;

    if(framesPerBuffer!=FramesPerBuffer)
    {
       if(bufferLenDiscrepancy==0)
       {
          printf("Discrepancy between the real framesPerBuffer (%d) and the projected one (%d).\n",
             framesPerBuffer, FramesPerBuffer);
          printf("Stop the program and correct the settings.\n");
          bufferLenDiscrepancy=1;
       }
       return paContinue;    
    }
    
    if( inputBuffer == NULL )
    {
        for( i=0; i<framesPerBuffer; i++ )
        {
            *out++ = 0;  /* left - silent */
            *out++ = 0;  /* right - silent */
        }
        gNumNoInputs += 1;
    }
    else
    {
        err = PitchBend(in, out);
        if(StereoMode==2)
           for( i=0; i<framesPerBuffer; i++ )
           {
              *out=*(out+1);
              out+=2;
           }
        else
        if(StereoMode==1)
        {
           out++;
           for( i=0; i<framesPerBuffer; i++ )
           {
              *out=*(out-1);
              out+=2;
           }
        }
    }
    
    return paContinue;
}


int SaveConfigFile()
{
   int i;
   FILE *config = fopen("pitchbenderconfig.txt", "wt");
   if(config==NULL)
   {
      printf("Could not open pitchbenderconfig.txt for writing\n");
      return 1;
   }
   for(i=0;i<inputfilelines;i++)
   {
      if(i==DefaultInputPos)
         fprintf(config, "DefaultInput: %d\n", DefaultInput); 
      else
      if(i==DefaultOutputPos)
         fprintf(config, "DefaultOutput: %d\n", DefaultOutput); 
      else
         fprintf(config, "%s", inputfilecontent[i]);
   }
   fclose(config);
   return 0;
}

int LoadConfigFile()
{
   int res, lines=0, files=0;
   char line[LINELEN];
   char field[LINELEN];
   char value[LINELEN];
   FILE *config = fopen("pitchbenderconfig.txt", "rt");
   if(config==NULL)
   {
      printf("Could not load pitchbenderconfig.txt\n");
      return 1;
   }
   while(!feof(config))
   {
      fgets(line, LINELEN, config);
      if(feof(config))
         break;
      res = sscanf(line, "%s %s", field, value);
      if(res==2&&strcmp(field, "File:")==0)
         files++;
      lines++;
   }
   rewind(config);
   inputfilecontent=(char**)malloc(lines*sizeof(char*));
   inputfilelines=lines;
   lines=0;
   InputFiles = (char**)malloc(files*sizeof(char*));
    while(!feof(config))
   {
      fgets(line, LINELEN, config);
      if(feof(config))
         break;
       inputfilecontent[lines]=(char*)malloc(strlen(line)+1);
       strcpy(inputfilecontent[lines++], line);
       if(line[0]=='#')
         continue;
      res = sscanf(line, "%s %s", field, value);
      if(res==0||res==EOF)
         continue;
      if(res!=2)
      {
         printf("Error in line: %s, result: %d ", line, res);
         fclose(config);
         return 1;
      }
      if(strcmp(field, "DefaultInput:")==0)
      {
         DefaultInputPos = lines-1;
         DefaultInput = atoi(value);
      }
      else
      if(strcmp(field, "DefaultOutput:")==0)
      {
         DefaultOutputPos = lines-1;
         DefaultOutput = atoi(value);
      }
      else
      if(strcmp(field, "InfiniteLoop:")==0)
      {
         if(value[0]=='Y'||value[0]=='y')
            InfiniteLoop = 1;
         else
            InfiniteLoop = 0;
      }
      else
      if(strcmp(field, "Stereo:")==0)
      {
         if(strcmp(value, "left-to-both")==0)
            StereoMode = 1;
         else
         if(strcmp(value, "right-to-both")==0)
            StereoMode = 2;
         else
            StereoMode = 0;
      }
      else
      if(strcmp(field, "DefaultKeyCode:")==0)
         DefaultKeyCode = atoi(value);
      else
      if(strcmp(field, "DefaultKeyName:")==0)
      {
         DefaultKeyName = (char*)malloc(strlen(value)+1);
         strcpy(DefaultKeyName, value);
      }
      else
      if(strcmp(field, "File:")==0)
      {
         InputFiles[InputFileCount] = (char*)malloc(strlen(value)+1);
         strcpy(InputFiles[InputFileCount++], value);
      }
      else
      {
         printf("Error in line: %s", line);
         fclose(config);
         return 1;
      }
   }
   if(InputFileCount==0)
   {
      printf("No input file specified in pitchbenderconfig.txt\n");
      return 1;
   }
   fclose(config);
   return 0;
}

void GetDevices(const PaDeviceInfo *deviceInfo, int numDevices, int *numinputdevices, int *numoutputdevices, char *inputDevices, char *outputDevices)
{
    int i;
    char *inputDevicesPtr = inputDevices;
    char *outputDevicesPtr = outputDevices;
    printf( "Number of devices = %d\n", numDevices );
    for( i=0; i<numDevices; i++ )
    {
        deviceInfo = Pa_GetDeviceInfo( i );
        printf( "device #%d: ", i );
        printf( "%s, ", deviceInfo->name );
        printf( "via %s\n",  Pa_GetHostApiInfo( deviceInfo->hostApi )->name );
        printf( "Max inputs = %d", deviceInfo->maxInputChannels  );
        printf( ", Max outputs = %d\n", deviceInfo->maxOutputChannels  );
        if(deviceInfo->maxInputChannels>0&&*numinputdevices<10)
        {
            printf( "Default low input latency   = %8.4f\n", deviceInfo->defaultLowInputLatency  );
            printf( "Default high input latency  = %8.4f\n", deviceInfo->defaultHighInputLatency  );
            *inputDevicesPtr++ = i+'0';
            (*numinputdevices)++;
        }
        if(deviceInfo->maxOutputChannels>0&&*numoutputdevices<10)
        {
            printf( "Default low output latency  = %8.4f\n", deviceInfo->defaultLowOutputLatency  );
            printf( "Default high output latency = %8.4f\n", deviceInfo->defaultHighOutputLatency  );
            *outputDevicesPtr++ = i+'0';
            (*numoutputdevices)++;
        }
        printf("\n");
    }
    *inputDevicesPtr = '\0';
    *outputDevicesPtr = '\0';
}

/*******************************************************************/
int main();
int main()
{
    PaStreamParameters inputParameters, outputParameters;
    int     numDevices, numinputdevices=0, numoutputdevices=0;
    char InputDevices[10], OutputDevices[10];
    const   PaDeviceInfo *deviceInfo;
    PaStream *stream = NULL;
    PaError err;
    
    const PaDeviceInfo* inputInfo;
    const PaDeviceInfo* outputInfo;
    int i, inputDeviceIndex, outputDeviceIndex;
    int numBytes;
    int numChannels;
    long minBufferSizeFrames;
    long maxBufferSizeFrames; 
    long preferredBufferSizeFrames; 
    long granularity;

    printf("pitchbender\n");
    printf("Andreja Andric 2018\n"); 
    fflush(stdout);
        
    if(LoadConfigFile())
       goto error2;

    printf("sizeof(int) = %lu\n", sizeof(int)); fflush(stdout);
    printf("sizeof(long) = %lu\n", sizeof(long)); fflush(stdout);

    err = Pa_Initialize();
    if( err != paNoError ) goto error;

    numDevices = Pa_GetDeviceCount();
    if( numDevices <= 0 )
    {
        printf( "ERROR: Pa_GetDeviceCount returned 0x%x\n", numDevices );
        err = numDevices;
        goto error;
    }
    
    /* Note that Pa_GetDefaultInputDevice() and Pa_GetDefaultOutputDevice() will only return ASIO devices 
       because PortAudio has been built for ASIO only. */
    GetDevices(deviceInfo, numDevices, &numinputdevices, &numoutputdevices, InputDevices, OutputDevices);
    
    if(numinputdevices>1)
    {
       printf("Specify the index of the input and output device:\n");
       DefaultInput = ChoiceFromStringWithDefault(InputDevices, (char)DefaultKeyCode, DefaultKeyName, DefaultInput );
       DefaultOutput = DefaultInput;
       inputParameters.device = DefaultInput; 
       outputParameters.device = DefaultOutput;
    }
    else
    {
       inputParameters.device = Pa_GetDefaultInputDevice(); /* default input device */
       outputParameters.device = Pa_GetDefaultOutputDevice(); /* default output device */
    }
    
    SaveConfigFile();  

    err = PaAsio_ShowControlPanel( inputParameters.device, NULL );
    if( err != paNoError ) goto error;

    err = PaAsio_GetAvailableBufferSizes( inputParameters.device, &minBufferSizeFrames, 
       &maxBufferSizeFrames, &preferredBufferSizeFrames, &granularity);
    if( err != paNoError ) goto error;

    FramesPerBuffer = preferredBufferSizeFrames;
    printf("Buffer length: %d frames\n", FramesPerBuffer);
    printf("Minimum buffer length: %d frames\n", minBufferSizeFrames);
    printf("Maximum buffer length: %d frames\n", maxBufferSizeFrames);
    printf("Granularity: %d\n", granularity);
  
    printf( "Input device # %d.\n", inputParameters.device );
    inputInfo = Pa_GetDeviceInfo( inputParameters.device );

    printf( "Output device # %d.\n", outputParameters.device );
    outputInfo = Pa_GetDeviceInfo( outputParameters.device );

    numChannels = inputInfo->maxInputChannels < outputInfo->maxOutputChannels
            ? inputInfo->maxInputChannels : outputInfo->maxOutputChannels;
    printf( "Num channels = %d.\n", numChannels );

    inputParameters.channelCount = numChannels;
    inputParameters.sampleFormat = PA_SAMPLE_TYPE;
    inputParameters.suggestedLatency = inputInfo->defaultLowInputLatency ;
    inputParameters.hostApiSpecificStreamInfo = NULL;

    outputParameters.channelCount = numChannels;
    outputParameters.sampleFormat = PA_SAMPLE_TYPE;
    outputParameters.suggestedLatency = outputInfo->defaultLowOutputLatency;
    outputParameters.hostApiSpecificStreamInfo = NULL;

    /* -- setup -- */

    err = Pa_OpenStream(
              &stream,
              &inputParameters,
              &outputParameters,
              SAMPLE_RATE,
              paFramesPerBufferUnspecified,
              0, /* don't bother clipping them */
              pitchbendCallback,
              NULL );

    {
       const PaHostErrorInfo *err2 = Pa_GetLastHostErrorInfo();
       printf("Last host error: %s\n", err2->errorText);
    }

    if( err != paNoError ) goto error;

    for(i=0;i<InputFileCount;i++)
    {
        if(PitchBenderInitialize(1, 200, 
            numChannels*FramesPerBuffer*SAMPLE_SIZE, 
            SAMPLE_RATE, 
            numChannels*SAMPLE_SIZE, 
            InputFiles[i], 
            InfiniteLoop)!=0)
        {
            printf("PitchBend initialization failed\n");
            goto error;
        }

        if(WaitForKeyPressed((char)DefaultKeyCode, DefaultKeyName, 'x', "exit"))
           break;

        err = Pa_StartStream( stream );
        if( err != paNoError ) goto error;
        printf("Wire on, file %s.\n", InputFiles[i]); 
        fflush(stdout);

        if(WaitForKeyPressed((char)DefaultKeyCode, DefaultKeyName, 'r', "repeat"))
           i--;

        printf("Wire off.\n"); 
        fflush(stdout);

        err = Pa_StopStream( stream );
        PitchBenderClose();
        if( err != paNoError ) goto error;
    }
    printf("Finished. gNumNoInputs = %d\n", gNumNoInputs );

    Pa_Terminate();
    return 0;

error:
    if( stream ) {
       Pa_AbortStream( stream );
       Pa_CloseStream( stream );
    }
    Pa_Terminate();
    printf( "An error occured while using the portaudio stream\n" );
    printf( "Error number: %d\n", err );
    printf( "Error message: %s\n", Pa_GetErrorText( err ) );
    WaitForAnyKeyPressed();
    return -1;

error2:
    printf( "An error occured. The application will close.\n" );
    WaitForAnyKeyPressed();
    return -1;

}

